/* 
    qasn1 - Asn1BerNode - data structure class for qasn1 library
    Copyright(C) 2010 - Andrius da Costa Ribas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Asn1BerNode.h"

class Asn1BerNodePrivate
{
public:

    Asn1BerNodePrivate( Asn1BerNode* parent = 0 );
    
    // related Nodes
    Asn1BerNode* m_parent;
    Asn1BerNode* m_firstChild;
    Asn1BerNode* m_next;
    Asn1BerNode* m_prev;
    
    unsigned int m_tag; // tag number
    QByteArray m_val; // value inside the node
    
    // tagClass remains as char because of calculation of the value in Parser class
    // as it is not possible assign an integer value into enumerated type in C++
    char m_tagClass; // class of this tag
    
    bool m_leaf; // whether it is a primitive value
    bool m_knownLen; // whether it's length is determined
    
    QString m_error;
};

Asn1BerNodePrivate::Asn1BerNodePrivate( Asn1BerNode* parent )
{
    m_prev = 0;
    m_next = 0;
    m_parent = parent;
    m_firstChild = 0;
    m_leaf = true;
    m_knownLen = false;
    m_error = "";
}

Asn1BerNode::Asn1BerNode() : d(new Asn1BerNodePrivate)
{
}

Asn1BerNode::Asn1BerNode( Asn1BerNode* parent ) : d(new Asn1BerNodePrivate(parent) )
{
    if ( parent != 0 )
    {
        if ( !(parent->hasChild()) )
        {
            parent->setChild(this);
        }
    }
}

void Asn1BerNode::setError(const QString& error)
{
    d->m_error = error;
}

bool Asn1BerNode::isValid()
{
    return d->m_error.isEmpty();
}

QString Asn1BerNode::error() const
{
    return d->m_error;
}

Asn1BerNode* Asn1BerNode::last()
{
    if ( d->m_next != 0 )
    {
        return d->m_next->last();
    }
    if ( d->m_firstChild != 0 )
    {
        return d->m_firstChild->last();
    }
    return this;
}

Asn1BerNode::~Asn1BerNode()
{
    if ( d->m_prev != 0 )
    {
        Q_ASSERT( d->m_prev != d->m_next);
        d->m_prev->setNext(d->m_next);
    }
    if ( d->m_next != 0 )
    {
        d->m_next->setPrev(d->m_prev);
        // in case of deleting first child:
        if ( d->m_parent != 0 )
        {
            d->m_next->setParent(d->m_parent);
            d->m_parent->setChild(d->m_next);
        }
    }
    else
    {
        // in case of deleting only child:
        if ( d->m_parent != 0 ) // only first child has direct link to parent
        {
            d->m_parent->setChild(0);
        }
    }
    // now delete all children
    if ( d->m_firstChild != 0 )
    {
        while ( d->m_firstChild->hasNext() )
        {
            delete (d->m_firstChild->next());
        }
        delete d->m_firstChild;
    }
    delete d;
}

Asn1BerNode::Asn1BerNode( const Asn1BerNode& other )
{
    Asn1BerNode* childPointer;
    Asn1BerNode* origPointer;
    
    setTag(other.tag());
    setClass(other.tagClass());
    if ( other.hasDefinedLength() )
    {
        setDefinedLength(true);
    }
    if ( other.isCodingComplex() )
    {
        setCodingComplex();
        // cloning all children
        origPointer = other.firstChild();
        setChild(origPointer->clone());
        childPointer = firstChild();
        while ( origPointer != 0)
        {
            origPointer = origPointer->next();
            if ( origPointer != 0 )
            {
                childPointer->setNext(origPointer->clone());
                childPointer->next()->setPrev(childPointer);
            }
        }
    }
    else
    {
        setVal(other.val());
    }
}

Asn1BerNode* Asn1BerNode::clone()
{
    Asn1BerNode* clonePointer;
    Asn1BerNode* childPointer;
    Asn1BerNode* origPointer;

    origPointer = this;
    clonePointer = new Asn1BerNode();
    
    clonePointer->setTag(d->m_tag);
    clonePointer->setClass(d->m_tagClass);
    if ( d->m_knownLen )
    {
        clonePointer->setDefinedLength(true);
    }
    if ( !(d->m_leaf) )
    {
        clonePointer->setCodingComplex();
        // cloning all children
        origPointer = origPointer->firstChild();
        clonePointer->setChild(origPointer->clone());
        childPointer = clonePointer->firstChild();
        while ( origPointer != 0)
        {
            origPointer = origPointer->next();
            if ( origPointer != 0 )
            {
                childPointer->setNext(origPointer->clone());
                childPointer->next()->setPrev(childPointer);
            }
        }
    }
    else
    {
        clonePointer->setVal(d->m_val);
    }
    return clonePointer;
}

void Asn1BerNode::insertBefore(Asn1BerNode* other)
{
    Q_ASSERT(other != 0);
    
    // if inserting first child then need to rearrange parent links
    if ( d->m_parent != 0 )
    {
        other->setParent(d->m_parent);
        d->m_parent->setChild(other);
        d->m_parent = 0;
    }
    
    if ( d->m_prev != 0 )
    {
        d->m_prev->setNext(other);
    }
    other->setPrev(d->m_prev);
    setPrev(other);
    other->setNext(this);
}

void Asn1BerNode::insertAfter(Asn1BerNode* other)
{
    Q_ASSERT(other != 0);
    
    if ( d->m_next != 0 )
    {
        d->m_next->setPrev(other);
    }
    other->setNext(d->m_next);
    setNext(other);
    other->setPrev(this);
}

void Asn1BerNode::setChild( Asn1BerNode* N )
{
    d->m_firstChild = N;
}

void Asn1BerNode::setParent( Asn1BerNode* N )
{
    d->m_parent = N;
}

void Asn1BerNode::setNext( Asn1BerNode* N )
{
    d->m_next = N;
}

void Asn1BerNode::setPrev( Asn1BerNode* N )
{
    d->m_prev = N;
}


bool Asn1BerNode::hasParent () const
{
    if ( hasPrev() )
    {
        return d->m_prev->hasParent();
    }
    return ( d->m_parent != 0 );
}

bool Asn1BerNode::hasChild () const
{
    return ( d->m_firstChild != 0 );
}

bool Asn1BerNode::hasNext () const
{
    return ( d->m_next != 0 );
}

bool Asn1BerNode::hasPrev () const
{
    return ( d->m_prev != 0 );
}

Asn1BerNode* Asn1BerNode::parent () const
{
    if ( hasPrev() )
    {
        return d->m_prev->parent();
    }
    return d->m_parent;
}

Asn1BerNode* Asn1BerNode::firstChild () const
{
    return d->m_firstChild;
}

Asn1BerNode* Asn1BerNode::next () const
{
    return d->m_next;
}

Asn1BerNode* Asn1BerNode::prev () const
{
    return d->m_prev;
}

unsigned int Asn1BerNode::tag () const
{
    return d->m_tag;
}

unsigned int  Asn1BerNode::len () const
{
    Q_ASSERT( hasDefinedLength() );
    if ( isCodingPrimitive() )
    {
        return d->m_val.size();
    }
    else
    {
        quint64 count = 0;
        Asn1BerNode* iterator;
        iterator = firstChild();
        while ( iterator != 0 )
        {
            count += iterator->rawSize();
            iterator = iterator->next();
        }
        return count;
    }
}

unsigned int Asn1BerNode::rawSize() const
{
    // basically, this does the same of serializer but instead of IO it counts how many bytes are to be used
    unsigned int size;
    unsigned int length;
    unsigned int tagNumber;
    
    // first the easy part, the Value:
    if ( isCodingPrimitive() )
    {
        size = d->m_val.size();
    }
    else
    {
        size = 0;
        Asn1BerNode* iterator;
        iterator = firstChild();
        while ( iterator != 0 )
        {
            size += iterator->rawSize();
            iterator = iterator->next();
        }
    }
    
    // now tag:
    tagNumber = d->m_tag;
    if ( tagNumber < BIG_TAG_IND )
    {
        ++size;
    }
    else
    {
        ++size;
        do
        {
            ++size;
            tagNumber /= 0x80;
        }while(tagNumber != 0);
    }
    
    // finally, length part:
    if ( !(hasDefinedLength()) )
    {
        ++size;
    }
    else
    {
        length = len();
        if ( length < BIG_LEN_IND )
        {
            ++size;
        }
        else
        {
            do
            {
                ++size;
                length /= 0x100;
            }while(length != 0);
            ++size;
        }
    }
    return size;
}

QByteArray Asn1BerNode::val () const
{
    return d->m_val;
}

void Asn1BerNode::setTag (unsigned int tag)
{
    d->m_tag = tag;
}

void Asn1BerNode::setVal (QByteArray val)
{
    d->m_val = val;
    d->m_knownLen = true;
}

void Asn1BerNode::setDefinedLength (bool defined)
{
    d->m_knownLen = defined;
}

bool Asn1BerNode::isClassUniversal () const
{
    return ( (unsigned char) d->m_tagClass == CLASS_UNIVERSAL );
}

bool Asn1BerNode::isClassApplication () const
{
    return ( (unsigned char) d->m_tagClass == CLASS_APPLICATION );
}

bool Asn1BerNode::isClassContext () const
{
    return ( (unsigned char) d->m_tagClass == CLASS_CONTEXT );
}

bool Asn1BerNode::isClassPrivate () const
{
    return ( (unsigned char) d->m_tagClass == CLASS_PRIVATE );
}

char Asn1BerNode::tagClass() const
{
    return d->m_tagClass;
}

bool Asn1BerNode::isCodingComplex () const
{
    return ( !(d->m_leaf) );
}

bool Asn1BerNode::isCodingPrimitive () const
{
    return ( d->m_leaf );
}

Asn1BerNode* Asn1BerNode::first ()
{
    if ( hasParent() )
    {
        return parent()->first();
    }
    if ( hasPrev() )
    {
        return prev()->first();
    }
    return this;
}

void Asn1BerNode::setClass( char tagClass )
{
    d->m_tagClass = tagClass;
}

void Asn1BerNode::setCodingPrimitive()
{
    while ( hasChild() )
    {
        delete firstChild();
    }
    d->m_leaf = true;
}

void Asn1BerNode::setCodingComplex()
{
    d->m_leaf = false;
}

bool Asn1BerNode::hasDefinedLength() const
{
    return ( d->m_knownLen );
}
